# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2007 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://posterity.edgewall.org/wiki/License.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://posterity.edgewall.org/log/.

from datetime import datetime, date

from babel.dates import format_date, format_time
from babel.support import Format
import formencode
from genshi.core import START, END, TEXT, QName, Attrs
from genshi.filters import HTMLFormFiller, Translator
from genshi.template import TemplateLoader
from paste.registry import StackedObjectProxy
from paste.util.multidict import MultiDict, UnicodeMultiDict
from paste.wsgiwrappers import WSGIRequest, environ_getter
from paste.wsgiwrappers import WSGIResponse as Response
import pkg_resources

from posterity import app, instance, __version__, request, response, session, log
from posterity.util import localtz
from posterity.util.compat import partial
from posterity.util.i18n import gettext, ngettext, gettext_noop, locale


def template_loaded(template):
    template.filters.insert(0, Translator(gettext))

tmpl_dir = pkg_resources.resource_filename('posterity', 'templates')
loader = TemplateLoader([tmpl_dir], auto_reload=True, 
                        variable_lookup='strict',
                        callback=template_loaded)


class Request(WSGIRequest):
    """A paste.WSGIRequest subclass with some extra features
    
    Additional features:
    * Genshi template rendering
    * req.params include wsgiorg.routing_args values
    """
    charset = 'utf-8'
    
    def __init__(self, environ):
        super(Request, self).__init__(environ)
        self.flash_msg = self.cookies.get('posterity-flash')
        if self.flash_msg:
            response.cookies['posterity-flash'] = ''
            response.cookies['posterity-flash']['path'] = '/'
            response.cookies['posterity-flash']['Max-Age'] = 0

    @property
    def formvars(self):
        """FormEncode validated form values"""
        return self.environ.get('posterity.formvars', {})


def flash(msg, type='info'):
    """Send a flash message.
    
    Flash messages are messages that are displayed on the next page
    viewed by the user. These messages are useful when the controller
    needs to redirect a user to a different page and likes to insert
    a message on that page.
    """
    response.cookies['posterity-flash'] = msg
    response.cookies['posterity-flash']['path'] = '/'

def redirect(start_response, location):
    response.status_code = 303
    response.headers['Location'] = location
    return response(request.environ, start_response)

def send_template(start_response, filename, data={}):
    """Renders a genshi template

    Forms will automatically be populated with form values and validation
    error messages if available.
    """
    if not data:
        data = {}
    tmpl = loader.load(filename)
    data.update(_default_context_data)
    data.setdefault('account', session.account)
    data.setdefault('form_errors', 
                    request.environ.get('posterity.errmsgs', {}))

    data.setdefault('format', Format(locale, localtz))
    data.setdefault('format_timedelta', 
                    partial(format_timedelta, locale=locale, tzinfo=localtz))

    stream = tmpl.generate(environ=request.environ, **data)
    if 'posterity.errvars' in request.environ:
        filler = HTMLFormFiller(data=request.environ['posterity.errvars'])
        stream = stream | filler
    form_token = request.environ['session'].cookie
    stream = stream | FormRewriteFilter() | CSRFFilter(form_token)
    data = stream.render('xhtml')
    response.status_code = 200
    response.headers['Content-type'] = 'text/html; charset=utf-8'
    response.content = data
    return response(request.environ, start_response)


class FormRewriteFilter(object):
    """Rewrite PUT/DELETE forms into POST forms.

    Most browsers only support form actions 'GET' and 'POST'.
    This filter works around this problem by setting the action attribute
    to 'POST' and inserting an extra hidden variable storing the actual
    http method. This filter should be used together with the
    `FormRewriteMiddleware`.
    
    >>> from genshi.input import HTML
    >>> stream = HTML('<form method="put"/>')
    >>> stream = stream | FormRewriteFilter()
    >>> stream.render()
    '<form method="post"><input type="hidden" name="_method" value="put"/></form>'
    """
    def __call__(self, stream):
        for kind, data, pos in stream:
            if kind is START:
                tag, attrib = data
                if tag == 'form':
                    method = attrib.get('method', 'get').lower()
                    if  method not in ('get', 'post'):
                        attrib = ((attrib - 'method') | \
                                  [(QName('method'), 'post')])
                        data = tag, attrib
                        yield kind, data, pos
                        # Store the real method in a hidden variable
                        attrib = Attrs([(QName('type'), 'hidden'),
                                        (QName('name'), '_method'), 
                                        (QName('value'), method)])
                        yield START, (QName('input'), attrib), -1
                        yield END, QName('input'), -1
                        continue
            yield kind, data, pos


class CSRFFilter(object):
    """Cross Site Request Forgery protection filter
    
    Inserts a shared secret hidden variable into all non GET-forms.
    CSRF attacks can be detected by validating this shared secret 
    when non-GET forms are submitted.
    
    >>> from genshi.input import HTML
    >>> stream = HTML('<form method="post"/>')
    >>> stream = stream | CSRFFilter('shared_secret')
    >>> stream.render()
    '<form method="post"><input type="hidden" name="_form_token" value="shared_secret"/></form>'
    """

    def __init__(self, form_token):
        self._form_token = form_token

    def __call__(self, stream):
        for kind, data, pos in stream:
            if kind is START:
                tag, attrib = data
                if tag == 'form':
                    method = attrib.get('method', 'get').lower()
                    if  method != 'get':
                        yield kind, data, pos
                        # Store the real method in a hidden variable
                        attrib = Attrs([(QName('type'), 'hidden'),
                                        (QName('name'), '_form_token'), 
                                        (QName('value'), self._form_token)])
                        yield START, (QName('input'), attrib), -1
                        yield END, QName('input'), -1
                        continue
            yield kind, data, pos


class ValidationState(dict):
    pass


def validate(schema, on_error=None):
    """A form validating decorator
    
    POST requests are validated against the specified `FormEncode` schema.
    On success the converted form variables are stored in the wsgi environment
    under the key "posterity.formvars". These values can also be accessed 
    from `req.formvars`.
    
    If the validation fail, the original messages and the error messages
    are stored in the wsgi environment and the request is re-run as a 
    GET-request. Forms on the resulting page will be filled with the values
    stored in the environment.
    """
    def wrapper(func):
        def wrapper2(self, environ, start_response):
            if environ['REQUEST_METHOD'] == 'GET':
                return func(self, environ, start_response)
            try:
                state = ValidationState()
                state.environ = environ
                state.account = environ['session'].account
                environ['posterity.formvars'] = schema().to_python(request.POST,
                                                                   state)
            except formencode.Invalid, e:
                # Keep track of (raw) form values and errors messages
                # So we can re-run the request as a GET request and fill the
                # form with values and error messages.
                environ['posterity.errvars'] = request.POST
                err_dict = e.unpack_errors()
                if not isinstance(err_dict, dict):
                    err_dict = dict(general=err_dict)
                environ['posterity.errmsgs'] = err_dict
                environ['REQUEST_METHOD'] = 'GET'
                environ['CONTENT_TYPE'] = ''
                environ['CONTENT_LENGTH'] = ''
                log.debug('Validation failed: %s' % e)
                if on_error:
                    environ['PATH_INFO'] = on_error
                else:
                    environ['PATH_INFO'] = environ['SCRIPT_NAME']
                    environ['SCRIPT_NAME'] = ''
                del environ['paste.parsed_formvars']
                return app(environ, start_response)
            return func(self, environ, start_response)
        return wrapper2
    return wrapper


def format_timedelta(ts, locale, tzinfo=None):
    today = datetime.utcnow().date()
    if ts.date() == today:
        return format_time(ts, locale=locale, tzinfo=tzinfo)
    else:
        return format_date(ts, 'long', locale=locale)

def shorten(text, max):
    if len(text) < max:
        return text
    else:
        return text[:max]+'...'

_default_context_data = {
    '__version__': __version__,
    'app': app,
    'instance': instance,
    'session': session,
    'request': request,
    'response': response,
    'gettext': gettext,
    'ngettext': ngettext,
    'gettext_noop': gettext_noop,
    'shorten': shorten
}
